/*
 * proxy_server.c
 *
 *  Created on: 2010-5-16
 *      Author: cai
 */

#include <config.h>

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <netinet/in.h>
#include <gio/gio.h>

#include "i18n.h"
#include "proxy_session.h"
#include "proxy_server.h"
#include "g_socket_source.h"
#ifndef _WIN32
#include "sd-daemon.h"
#endif

static int	systemd_mode = FALSE;

struct ProxyServer{
	GObject parent;
	GList * listener;
};

struct ProxyServerClass{
	GObjectClass parent_class;
};

static void proxy_server_class_init(ProxyServerClass*);
static void proxy_server_dispose(ProxyServer*);
static void proxy_server_finalize(ProxyServer*);
static void proxy_server_init(ProxyServer*);

G_DEFINE_TYPE(ProxyServer,proxy_server,G_TYPE_OBJECT);

static void g_cclosure_marshal_VOID__LONG_STRING(GClosure *closure,
		GValue *return_value, guint n_param_values, const GValue *param_values,
		gpointer invocation_hint, gpointer marshal_data);

static void proxy_server_class_init(ProxyServerClass* klass)
{
	g_signal_new("log",G_OBJECT_CLASS_TYPE(klass),G_SIGNAL_ACTION|G_SIGNAL_RUN_LAST,0,0,0,g_cclosure_marshal_VOID__LONG_STRING,G_TYPE_NONE,2,G_TYPE_LONG,G_TYPE_STRING);
	G_OBJECT_CLASS(klass)->finalize = (gpointer) proxy_server_finalize;
	G_OBJECT_CLASS(klass)->dispose = (gpointer) proxy_server_dispose;
}

static void proxy_server_dispose(ProxyServer* server)
{
	G_OBJECT_CLASS(proxy_server_parent_class)->dispose(G_OBJECT(server));
}

static void proxy_server_finalize(ProxyServer*server)
{
	G_OBJECT_CLASS(proxy_server_parent_class)->finalize(G_OBJECT(server));
}

static void proxy_server_init(ProxyServer* obj)
{

}

ProxyServer * proxy_server_new()
{
	return (ProxyServer*)g_object_new(PROXY_TYPE_SERVER,NULL);
}

gboolean proxy_server_is_by_systemd(ProxyServer*proxy)
{
#ifndef _WIN32
	GSocket * sock = NULL;

	//check if start in systemd mode
	int fds = sd_listen_fds(0);

	if(fds > 0)
	{
		for (int var = 0; var < fds; ++var)
		{
			sock = g_socket_new_from_fd(var+SD_LISTEN_FDS_START,NULL);

			g_socket_set_blocking(sock,FALSE);

			proxy->listener = g_list_append(proxy->listener, sock);
		}
		systemd_mode = TRUE;
	}
	return systemd_mode;
#else
	return 0;
#endif
}

gboolean proxy_server_add_port(ProxyServer * proxy, gushort port,GError ** err)
{
	GSocket * sock = NULL;
	gboolean status = 0 ;

	int one=1;

	do
	{
		sock =  g_socket_new(G_SOCKET_FAMILY_IPV4, G_SOCKET_TYPE_STREAM,G_SOCKET_PROTOCOL_DEFAULT, err);
		if (!sock)
			break;

		setsockopt(g_socket_get_fd(sock), SOL_IP, IP_TRANSPARENT, &one, sizeof(one));

		GInetAddress *  address = g_inet_address_new_any(G_SOCKET_FAMILY_IPV4);

		GSocketAddress * addr = g_inet_socket_address_new(address,port);

		if(!(status = g_socket_bind(sock,addr,TRUE,err)))
		{
			break;
		}

		if(!(status = g_socket_listen(sock,err)))
		{
			break;
		}

//		sock = g_socket_new_from_fd(dup(g_socket_get_fd(sock)),NULL);

		g_object_unref(address);
		g_object_unref(addr);

	} while (0);
	if(!status && sock)
	{
		g_object_unref(sock);
	}
	else
	{
		g_socket_set_blocking(sock,FALSE);

		proxy->listener = g_list_append(proxy->listener, sock);
	}
	return status || systemd_mode;
}

static	gboolean new_http_request(GSocket *socket,GIOCondition condition,ProxyServer * server)
{
	int one=1;

	g_debug("new_http_request");

	GSocket * client = 	g_socket_accept(socket,0,NULL);

	if(!client)
	{
		return TRUE;
	}

	Session * session = proxy_session_new(client,server);

	if(session)
	{
		g_debug("accept new client, assign to session %p",session);
		g_object_unref(session);
	}
	return TRUE;
}

static	void start_accpet(gpointer data, ProxyServer *server)
{
	g_socket_add_watch(data,G_IO_IN,999,(GSocketSourceFunc)new_http_request,server);
}

gboolean proxy_server_start(ProxyServer * server)
{
	g_list_foreach(server->listener,(GFunc)start_accpet,server);
}

static void g_cclosure_marshal_VOID__LONG_STRING(GClosure *closure,
		GValue *return_value, guint n_param_values, const GValue *param_values,
		gpointer invocation_hint, gpointer marshal_data)
{

	  typedef void (*GMarshalFunc_VOID__LONG_STRING) (gpointer     data1,
	                                              gint         arg_1,
	                                              gchar   *    arg_2,
	                                              gpointer     data2);
	  register GMarshalFunc_VOID__LONG_STRING callback;
	  register GCClosure *cc = (GCClosure*) closure;
	  register gpointer data1, data2;

	  g_return_if_fail (n_param_values == 3);

	  if (G_CCLOSURE_SWAP_DATA (closure))
	    {
	      data1 = closure->data;
	      data2 = g_value_peek_pointer (param_values + 0);
	    }
	  else
	    {
	      data1 = g_value_peek_pointer (param_values + 0);
	      data2 = closure->data;
	    }
	  callback = (GMarshalFunc_VOID__LONG_STRING) (marshal_data ? marshal_data : cc->callback);

	  callback (data1,
	            g_value_get_long(param_values + 1),
	            (char*)g_value_get_string(param_values + 2),
	            data2);
}
